//===========================================================================
//
//  Code written for PSoC6 on a Pioneer BLE board
//
//  CRYSTALFONTZ CFAF240240A1-013T 240X240 SPI COLOR 1.3" TFT
//  ref: https://www.crystalfontz.com/product/cfaf240240a1013t
//
//  This code uses 4-wire SPI mode.
//
// The controller is a Sitronix ST7789H2
//   http://www.crystalfontz.com/controllers/Sitronix/ST7789H2
//
//  2018-10-19 Brent A. Crosby
//===========================================================================
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//============================================================================
//
// Display is Crystalfontz  CFAF240240A1-013T 240X240 SPI COLOR 1.3" TFT
//   https://www.crystalfontz.com/product/cfaf240240a1013t
//
//============================================================================
#include "project.h"
#include "CFAF240240A_LCD.h"
#include "SPI_DMA_LCD.h"
#include "images.h"

//============================================================================
// Green channel of RGB LED on Poieer board
#define GREEN_LED_ON  Cy_GPIO_Write(GreenLED_PORT,GreenLED_NUM,0)
#define GREEN_LED_OFF Cy_GPIO_Write(GreenLED_PORT,GreenLED_NUM,1)
//============================================================================
//Some 16-bit 565 RRRRR GGG GGG BBBBB sample colors for fill
const uint16_t colors[8]=
  {
  0xFFFF, //white
  0x0000, //black
  0xF800, //red
  0x07E0, //green
  0x001F, //blue
  0xF800|0x07E0, //yellow
  0xF800|0x001F, //purple
  0x07E0|0x001F  //cyan
  };
//----------------------------------------------------------------------------
int main(void)
  {
  uint8_t
     what_to_show;
  what_to_show=0;
    
  //Set up the DMA from the framebuffer to the SPI
  Initialize_SPI_DMA_for_LCD();
       
  // Enable SPI master hardware
  Cy_SCB_SPI_Enable(LCD_SPI_HW);
  
  //Use non-DMA SPI transfers to initialize the display hardware
  LCD_Initialize();

  // Enable global interrupts.
  __enable_irq(); 
  
  for(;;)
    {
    //==== DRAWING HERE ======================================================
    //Put something more-or-less interesting into the framebuffer
    switch(what_to_show)
      {
      case 0:
        LCD_Flash_Image_to_Framebuffer(splash);
        break;
      case 1:
        LCD_Circle_Demo();
        break;
      case 2:
        LCD_Flash_Image_to_Framebuffer(bug);
        break;
      case 3:
        LCD_Line_Demo();
        break;
      case 4:
        LCD_Flash_Image_to_Framebuffer(watch);
        break;
      case 5:
        LCD_Flash_Image_to_Framebuffer(crayons);
        break;
      default:
        LCD_FillFrameBuffer(colors[what_to_show-6]);
      }  
    //Turn the green LED on to indicate we are doing an update
    GREEN_LED_ON;
    //==== UPDATE START ======================================================
    //Fire off the DMA to transfer the framebuffer to the LCD.
    Start_SPI_DMA_transfer_to_LCD();
    //Wait for the DMA to finish.
    while(SPI_DMA_LCD_IDLE != SPI_DMA_LCD_status)
      {
      //You could do something useful instead of waiting
      // for the 36.8mS @25MHz SPI Clock that this takes
      // 921600 bits / 0.0368 second = 25 Mbits/second         
      }
    //==== UPDATE END ========================================================
    //The transfer is done, turn off the green LED.
    GREEN_LED_OFF;

    //Give a bit to let the user see the image.
    if(what_to_show <= 5)
      //bitmap image
      CyDelay(3000);
    else
      //filled screen
      CyDelay(1000);

    //Pick some other thing for next time
    if(what_to_show<13)
      what_to_show++;
    else
      what_to_show=0;
    }
  }
//============================================================================
